@@ -1,14 +1,12 @@
 import os.path
+import ruamel.yaml as yaml
 import string
 from abc import ABCMeta, abstractmethod
 from collections import OrderedDict
 from copy import copy
 from datetime import datetime
-from itertools import chain
 from warnings import warn
 
-import ruamel.yaml as yaml
-
 from .catalog import SpecCatalog
 from .spec import DatasetSpec, GroupSpec
 from ..utils import docval, getargs, popargs, get_docval, call_docval_func
@@ -203,7 +201,8 @@ class YAMLSpecReader(SpecReader):
     def read_namespace(self, namespace_path):
         namespaces = None
         with open(namespace_path, 'r') as stream:
-            d = yaml.safe_load(stream)
+            yaml_obj = yaml.YAML(typ='safe', pure=True)
+            d = yaml_obj.load(stream)
             namespaces = d.get('namespaces')
             if namespaces is None:
                 raise ValueError("no 'namespaces' found in %s" % namespace_path)
@@ -212,7 +211,8 @@ class YAMLSpecReader(SpecReader):
     def read_spec(self, spec_path):
         specs = None
         with open(self.__get_spec_path(spec_path), 'r') as stream:
-            specs = yaml.safe_load(stream)
+            yaml_obj = yaml.YAML(typ='safe', pure=True)
+            specs = yaml_obj.load(stream)
             if not ('datasets' in specs or 'groups' in specs):
                 raise ValueError("no 'groups' or 'datasets' found in %s" % spec_path)
         return specs
@@ -290,7 +290,9 @@ class NamespaceCatalog:
         self.__namespaces[name] = namespace
         for dt in namespace.catalog.get_registered_types():
             source = namespace.catalog.get_spec_source_file(dt)
-            self.__loaded_specs.setdefault(source, list()).append(dt)
+            # do not add types that have already been loaded
+            # use dict with None values as ordered set because order of specs does matter
+            self.__loaded_specs.setdefault(source, dict()).update({dt: None})
 
     @docval({'name': 'name', 'type': str, 'doc': 'the name of this namespace'},
             returns="the SpecNamespace with the given name", rtype=SpecNamespace)
@@ -365,67 +367,85 @@ class NamespaceCatalog:
         ret = self.__loaded_specs.get(source)
         if ret is not None:
             ret = tuple(ret)
+        else:
+            ret = tuple()
         return ret
 
-    def __load_spec_file(self, reader, spec_source, catalog, dtypes=None, resolve=True):
+    def __load_spec_file(self, reader, spec_source, catalog, types_to_load=None, resolve=True):
         ret = self.__loaded_specs.get(spec_source)
         if ret is not None:
             raise ValueError("spec source '%s' already loaded" % spec_source)
 
         def __reg_spec(spec_cls, spec_dict):
-            parent_cls = GroupSpec if issubclass(spec_cls, GroupSpec) else DatasetSpec
-            dt_def = spec_dict.get(spec_cls.def_key(), spec_dict.get(parent_cls.def_key()))
+            dt_def = spec_dict.get(spec_cls.def_key())
             if dt_def is None:
-                msg = 'no %s or %s found in spec %s' % (spec_cls.def_key(), parent_cls.def_key(), spec_source)
+                msg = 'No data type def key found in spec %s' % spec_source
                 raise ValueError(msg)
-            if dtypes and dt_def not in dtypes:
+            if types_to_load and dt_def not in types_to_load:
                 return
             if resolve:
-                self.__resolve_includes(spec_dict, catalog)
+                self.__resolve_includes(spec_cls, spec_dict, catalog)
             spec_obj = spec_cls.build_spec(spec_dict)
             return catalog.auto_register(spec_obj, spec_source)
 
         if ret is None:
-            ret = list()
+            ret = dict()  # this is used as an ordered set -- values are all none
             d = reader.read_spec(spec_source)
             specs = d.get('datasets', list())
             for spec_dict in specs:
-                ret.extend(__reg_spec(self.__dataset_spec_cls, spec_dict))
+                self.__convert_spec_cls_keys(GroupSpec, self.__group_spec_cls, spec_dict)
+                temp_dict = {k: None for k in __reg_spec(self.__dataset_spec_cls, spec_dict)}
+                ret.update(temp_dict)
             specs = d.get('groups', list())
             for spec_dict in specs:
-                ret.extend(__reg_spec(self.__group_spec_cls, spec_dict))
+                self.__convert_spec_cls_keys(GroupSpec, self.__group_spec_cls, spec_dict)
+                temp_dict = {k: None for k in __reg_spec(self.__group_spec_cls, spec_dict)}
+                ret.update(temp_dict)
             self.__loaded_specs[spec_source] = ret
         return ret
 
-    def __resolve_includes(self, spec_dict, catalog):
+    def __convert_spec_cls_keys(self, parent_cls, spec_cls, spec_dict):
+        """Replace instances of data_type_def/inc in spec_dict with new values from spec_cls."""
+        # this is necessary because the def_key and inc_key may be different in each namespace
+        # NOTE: this does not handle more than one custom set of keys
+        if parent_cls.def_key() in spec_dict:
+            spec_dict[spec_cls.def_key()] = spec_dict.pop(parent_cls.def_key())
+        if parent_cls.inc_key() in spec_dict:
+            spec_dict[spec_cls.inc_key()] = spec_dict.pop(parent_cls.inc_key())
+
+    def __resolve_includes(self, spec_cls, spec_dict, catalog):
+        """Replace data type inc strings with the spec definition so the new spec is built with included fields.
         """
-            Pull in any attributes, datasets, or groups included
-        """
-        dt_inc = spec_dict.get(self.__group_spec_cls.inc_key())
-        dt_def = spec_dict.get(self.__group_spec_cls.def_key())
+        dt_def = spec_dict.get(spec_cls.def_key())
+        dt_inc = spec_dict.get(spec_cls.inc_key())
         if dt_inc is not None and dt_def is not None:
             parent_spec = catalog.get_spec(dt_inc)
             if parent_spec is None:
                 msg = "Cannot resolve include spec '%s' for type '%s'" % (dt_inc, dt_def)
                 raise ValueError(msg)
-            spec_dict[self.__group_spec_cls.inc_key()] = parent_spec
-        it = chain(spec_dict.get('groups', list()), spec_dict.get('datasets', list()))
-        for subspec_dict in it:
-            self.__resolve_includes(subspec_dict, catalog)
+            # replace the inc key value from string to the inc spec so that the spec can be updated with all of the
+            # attributes, datasets, groups, and links of the inc spec when spec_cls.build_spec(spec_dict) is called
+            spec_dict[spec_cls.inc_key()] = parent_spec
+        for subspec_dict in spec_dict.get('groups', list()):
+            self.__resolve_includes(self.__group_spec_cls, subspec_dict, catalog)
+        for subspec_dict in spec_dict.get('datasets', list()):
+            self.__resolve_includes(self.__dataset_spec_cls, subspec_dict, catalog)
 
-    def __load_namespace(self, namespace, reader, types_key, resolve=True):
+    def __load_namespace(self, namespace, reader, resolve=True):
         ns_name = namespace['name']
-        if ns_name in self.__namespaces:
+        if ns_name in self.__namespaces:  # pragma: no cover
             raise KeyError("namespace '%s' already exists" % ns_name)
         catalog = SpecCatalog()
         included_types = dict()
         for s in namespace['schema']:
+            # types_key may be different in each spec namespace, so check both the __spec_namespace_cls types key
+            # and the parent SpecNamespace types key. NOTE: this does not handle more than one custom types key
+            types_to_load = s.get(self.__spec_namespace_cls.types_key(), s.get(SpecNamespace.types_key()))
+            if types_to_load is not None:  # schema specifies specific types from 'source' or 'namespace'
+                types_to_load = set(types_to_load)
             if 'source' in s:
                 # read specs from file
-                dtypes = None
-                if types_key in s:
-                    dtypes = set(s[types_key])
-                self.__load_spec_file(reader, s['source'], catalog, dtypes=dtypes, resolve=resolve)
+                self.__load_spec_file(reader, s['source'], catalog, types_to_load=types_to_load, resolve=resolve)
                 self.__included_sources.setdefault(ns_name, list()).append(s['source'])
             elif 'namespace' in s:
                 # load specs from namespace
@@ -433,11 +453,9 @@ class NamespaceCatalog:
                     inc_ns = self.get_namespace(s['namespace'])
                 except KeyError as e:
                     raise ValueError("Could not load namespace '%s'" % s['namespace']) from e
-                if types_key in s:
-                    types = s[types_key]
-                else:
-                    types = inc_ns.get_registered_types()
-                for ndt in types:
+                if types_to_load is None:
+                    types_to_load = inc_ns.get_registered_types()  # load all types in namespace
+                for ndt in types_to_load:
                     spec = inc_ns.get_spec(ndt)
                     spec_file = inc_ns.catalog.get_spec_source_file(ndt)
                     if isinstance(spec, DatasetSpec):
@@ -445,7 +463,9 @@ class NamespaceCatalog:
                     else:
                         spec = self.group_spec_cls.build_spec(spec)
                     catalog.register_spec(spec, spec_file)
-                included_types[s['namespace']] = tuple(types)
+                included_types[s['namespace']] = tuple(types_to_load)
+            else:
+                raise ValueError("Spec '%s' schema must have either 'source' or 'namespace' key" % ns_name)
         # construct namespace
         ns = self.__spec_namespace_cls.build_namespace(catalog=catalog, **namespace)
         self.__namespaces[ns_name] = ns
@@ -475,7 +495,6 @@ class NamespaceCatalog:
         else:
             return ret
         namespaces = reader.read_namespace(namespace_path)
-        types_key = self.__spec_namespace_cls.types_key()
         to_load = list()
         for ns in namespaces:
             if ns['name'] in self.__namespaces:
@@ -487,6 +506,6 @@ class NamespaceCatalog:
                 to_load.append(ns)
         # now load specs into namespace
         for ns in to_load:
-            ret[ns['name']] = self.__load_namespace(ns, reader, types_key, resolve=resolve)
+            ret[ns['name']] = self.__load_namespace(ns, reader, resolve=resolve)
         self.__included_specs[ns_path_key] = ret
         return ret
