@@ -10,7 +10,7 @@ import numpy as np
 
 __macros = {
     'array_data': [np.ndarray, list, tuple, h5py.Dataset],
-    'scalar_data': [str, int, float, bytes],
+    'scalar_data': [str, int, float, bytes, bool],
     'data': []
 }
 
