@@ -2,11 +2,12 @@ import re
 from abc import ABCMeta, abstractmethod
 from copy import copy
 from itertools import chain
+from collections import defaultdict
 
 import numpy as np
 
 from .errors import Error, DtypeError, MissingError, MissingDataType, ShapeError, IllegalLinkError, IncorrectDataType
-from .errors import ExpectedArrayError
+from .errors import ExpectedArrayError, IncorrectQuantityError
 from ..build import GroupBuilder, DatasetBuilder, LinkBuilder, ReferenceBuilder, RegionBuilder
 from ..build.builders import BaseBuilder
 from ..spec import Spec, AttributeSpec, GroupSpec, DatasetSpec, RefSpec, LinkSpec
@@ -110,6 +111,8 @@ def get_type(data):
         if data.size == 0:
             raise EmptyArrayError()
         return get_type(data[0])
+    elif isinstance(data, np.bool_):
+        return 'bool'
     if not hasattr(data, '__len__'):
         return type(data).__name__
     else:
@@ -156,16 +159,16 @@ class ValidatorMap:
     def __init__(self, **kwargs):
         ns = getargs('namespace', kwargs)
         self.__ns = ns
-        tree = dict()
+        tree = defaultdict(list)
         types = ns.get_registered_types()
         self.__type_key = ns.get_spec(types[0]).type_key()
         for dt in types:
             spec = ns.get_spec(dt)
             parent = spec.data_type_inc
             child = spec.data_type_def
-            tree.setdefault(child, list())
+            tree[child] = list()
             if parent is not None:
-                tree.setdefault(parent, list()).append(child)
+                tree[parent].append(child)
         for t in tree:
             self.__rec(tree, t)
         self.__valid_types = dict()
@@ -184,12 +187,11 @@ class ValidatorMap:
             self.__valid_types[dt] = tuple(_list)
 
     def __rec(self, tree, node):
-        if isinstance(tree[node], tuple):
-            return tree[node]
-        sub_types = {node}
-        for child in tree[node]:
-            sub_types.update(self.__rec(tree, child))
-        tree[node] = tuple(sub_types)
+        if not isinstance(tree[node], tuple):
+            sub_types = {node}
+            for child in tree[node]:
+                sub_types.update(self.__rec(tree, child))
+            tree[node] = tuple(sub_types)
         return tree[node]
 
     @property
@@ -395,6 +397,12 @@ class DatasetValidator(BaseStorageValidator):
         return ret
 
 
+def _resolve_data_type(spec):
+    if isinstance(spec, LinkSpec):
+        return spec.target_type
+    return spec.data_type
+
+
 class GroupValidator(BaseStorageValidator):
     '''A class for validating GroupBuilders against GroupSpecs'''
 
@@ -402,78 +410,239 @@ class GroupValidator(BaseStorageValidator):
             {'name': 'validator_map', 'type': ValidatorMap, 'doc': 'the ValidatorMap to use during validation'})
     def __init__(self, **kwargs):
         call_docval_func(super().__init__, kwargs)
-        self.__include_dts = dict()
-        self.__dataset_validators = dict()
-        self.__group_validators = dict()
-        it = chain(self.spec.datasets, self.spec.groups)
-        for spec in it:
-            if spec.data_type_def is None:
-                if spec.data_type_inc is None:
-                    if isinstance(spec, GroupSpec):
-                        self.__group_validators[spec.name] = GroupValidator(spec, self.vmap)
-                    else:
-                        self.__dataset_validators[spec.name] = DatasetValidator(spec, self.vmap)
-                else:
-                    self.__include_dts[spec.data_type_inc] = spec
-            else:
-                self.__include_dts[spec.data_type_def] = spec
-
-        for spec in self.spec.links:
-            self.__include_dts[spec.data_type_inc] = spec
 
     @docval({"name": "builder", "type": GroupBuilder, "doc": "the builder to validate"},  # noqa: C901
             returns='a list of Errors', rtype=list)
     def validate(self, **kwargs):  # noqa: C901
         builder = getargs('builder', kwargs)
-        ret = super().validate(builder)
-        # get the data_types
-        data_types = dict()
-        for key, value in builder.items():
-            v_builder = value
-            if isinstance(v_builder, LinkBuilder):
-                v_builder = v_builder.builder
-            if isinstance(v_builder, BaseBuilder):
-                dt = v_builder.attributes.get(self.spec.type_key())
-                if dt is not None:
-                    data_types.setdefault(dt, list()).append(value)
-        for dt, inc_spec in self.__include_dts.items():
-            found = False
-            inc_name = inc_spec.name
-            for sub_val in self.vmap.valid_types(dt):
-                spec = sub_val.spec
-                sub_dt = spec.data_type_def
-                dt_builders = data_types.get(sub_dt)
-                if dt_builders is not None:
-                    if inc_name is not None:
-                        dt_builders = filter(lambda x: x.name == inc_name, dt_builders)
-                    for bldr in dt_builders:
-                        tmp = bldr
-                        if isinstance(bldr, LinkBuilder):
-                            if isinstance(inc_spec, LinkSpec) or inc_spec.linkable:
-                                tmp = bldr.builder
-                            else:
-                                ret.append(IllegalLinkError(self.get_spec_loc(inc_spec),
-                                                            location=self.get_builder_loc(tmp)))
-                        ret.extend(sub_val.validate(tmp))
-                        found = True
-            if not found and self.__include_dts[dt].required:
-                ret.append(MissingDataType(self.get_spec_loc(self.spec), dt,
-                                           location=self.get_builder_loc(builder), missing_dt_name=inc_name))
-        it = chain(self.__dataset_validators.items(),
-                   self.__group_validators.items())
-        for name, validator in it:
-            sub_builder = builder.get(name)
-            spec = validator.spec
-            if isinstance(sub_builder, LinkBuilder):
-                if spec.linkable:
-                    sub_builder = sub_builder.builder
-                else:
-                    ret.append(IllegalLinkError(self.get_spec_loc(spec), location=self.get_builder_loc(builder)))
-                    continue
-            if sub_builder is None:
-                if spec.required:
-                    ret.append(MissingError(self.get_spec_loc(spec), location=self.get_builder_loc(builder)))
-            else:
-                ret.extend(validator.validate(sub_builder))
 
-        return ret
+        errors = super().validate(builder)
+        errors.extend(self.__validate_children(builder))
+        return errors
+
+    def __validate_children(self, parent_builder):
+        """Validates the children of the group builder against the children in the spec.
+
+        Children are defined as datasets, groups, and links.
+
+        Validation works by first assigning builder children to spec children
+        in a many-to-one relationship using a SpecMatcher (this matching is
+        non-trivial due to inheritance, which is why it is isolated in a
+        separate class). Once the matching is complete, it is a
+        straightforward procedure for validating the set of matching builders
+        against each child spec.
+        """
+        spec_children = chain(self.spec.datasets, self.spec.groups, self.spec.links)
+        matcher = SpecMatcher(self.vmap, spec_children)
+
+        builder_children = chain(parent_builder.datasets.values(),
+                                 parent_builder.groups.values(),
+                                 parent_builder.links.values())
+        matcher.assign_to_specs(builder_children)
+
+        for child_spec, matched_builders in matcher.spec_matches:
+            yield from self.__validate_presence_and_quantity(child_spec, len(matched_builders), parent_builder)
+            for child_builder in matched_builders:
+                yield from self.__validate_child_builder(child_spec, child_builder, parent_builder)
+
+    def __validate_presence_and_quantity(self, child_spec, n_builders, parent_builder):
+        """Validate that at least one matching builder exists if the spec is
+        required and that the number of builders agrees with the spec quantity
+        """
+        if n_builders == 0 and child_spec.required:
+            yield self.__construct_missing_child_error(child_spec, parent_builder)
+        elif self.__incorrect_quantity(n_builders, child_spec):
+            yield self.__construct_incorrect_quantity_error(child_spec, parent_builder, n_builders)
+
+    def __construct_missing_child_error(self, child_spec, parent_builder):
+        """Returns either a MissingDataType or a MissingError depending on
+        whether or not a specific data type can be resolved from the spec
+        """
+        data_type = _resolve_data_type(child_spec)
+        builder_loc = self.get_builder_loc(parent_builder)
+        if data_type is not None:
+            name_of_erroneous = self.get_spec_loc(self.spec)
+            return MissingDataType(name_of_erroneous, data_type,
+                                   location=builder_loc, missing_dt_name=child_spec.name)
+        else:
+            name_of_erroneous = self.get_spec_loc(child_spec)
+            return MissingError(name_of_erroneous, location=builder_loc)
+
+    @staticmethod
+    def __incorrect_quantity(n_found, spec):
+        """Returns a boolean indicating whether the number of builder elements matches the specified quantity"""
+        if not spec.is_many() and n_found > 1:
+            return True
+        elif isinstance(spec.quantity, int) and n_found != spec.quantity:
+            return True
+        return False
+
+    def __construct_incorrect_quantity_error(self, child_spec, parent_builder, n_builders):
+        name_of_erroneous = self.get_spec_loc(self.spec)
+        data_type = _resolve_data_type(child_spec)
+        builder_loc = self.get_builder_loc(parent_builder)
+        return IncorrectQuantityError(name_of_erroneous, data_type, expected=child_spec.quantity,
+                                      received=n_builders, location=builder_loc)
+
+    def __validate_child_builder(self, child_spec, child_builder, parent_builder):
+        """Validate a child builder against a child spec considering links"""
+        if isinstance(child_builder, LinkBuilder):
+            if self.__cannot_be_link(child_spec):
+                yield self.__construct_illegal_link_error(child_spec, parent_builder)
+                return  # do not validate illegally linked objects
+            child_builder = child_builder.builder
+        child_validator = self.__get_child_validator(child_spec)
+        yield from child_validator.validate(child_builder)
+
+    def __construct_illegal_link_error(self, child_spec, parent_builder):
+        name_of_erroneous = self.get_spec_loc(child_spec)
+        builder_loc = self.get_builder_loc(parent_builder)
+        return IllegalLinkError(name_of_erroneous, location=builder_loc)
+
+    @staticmethod
+    def __cannot_be_link(spec):
+        return not isinstance(spec, LinkSpec) and not spec.linkable
+
+    def __get_child_validator(self, spec):
+        """Returns the appropriate validator for a child spec
+
+        If a specific data type can be resolved, the validator is acquired from
+        the ValidatorMap, otherwise a new Validator is created.
+        """
+        if _resolve_data_type(spec) is not None:
+            return self.vmap.get_validator(_resolve_data_type(spec))
+        elif isinstance(spec, GroupSpec):
+            return GroupValidator(spec, self.vmap)
+        elif isinstance(spec, DatasetSpec):
+            return DatasetValidator(spec, self.vmap)
+        else:
+            msg = "Unable to resolve a validator for spec %s" % spec
+            raise ValueError(msg)
+
+
+class SpecMatches:
+    """A utility class to hold a spec and the builders matched to it"""
+
+    def __init__(self, spec):
+        self.spec = spec
+        self.builders = list()
+
+    def add(self, builder):
+        self.builders.append(builder)
+
+
+class SpecMatcher:
+    """Matches a set of builders against a set of specs
+
+    This class is intended to isolate the task of choosing which spec a
+    builder should be validated against from the task of performing that
+    validation.
+    """
+
+    def __init__(self, vmap, specs):
+        self.vmap = vmap
+        self._spec_matches = [SpecMatches(spec) for spec in specs]
+        self._unmatched_builders = SpecMatches(None)
+
+    @property
+    def unmatched_builders(self):
+        """Returns the builders for which no matching spec was found
+
+        These builders can be considered superfluous, and will generate a
+        warning in the future.
+        """
+        return self._unmatched_builders.builders
+
+    @property
+    def spec_matches(self):
+        """Returns a list of tuples of: (spec, assigned builders)"""
+        return [(sm.spec, sm.builders) for sm in self._spec_matches]
+
+    def assign_to_specs(self, builders):
+        """Assigns a set of builders against a set of specs (many-to-one)
+
+        In the case that no matching spec is found, a builder will be
+        added to a list of unmatched builders.
+        """
+        for builder in builders:
+            spec_match = self._best_matching_spec(builder)
+            if spec_match is None:
+                self._unmatched_builders.add(builder)
+            else:
+                spec_match.add(builder)
+
+    def _best_matching_spec(self, builder):
+        """Finds the best matching spec for builder
+
+        The current algorithm is:
+        1. filter specs which meet the minimum requirements of consistent name
+            and data type
+        2. if more than one candidate meets the minimum requirements, find the
+            candidates which do not yet have a sufficient number of builders
+            assigned (based on the spec quantity)
+        3. return the first unsatisfied candidate if any, otherwise return the
+            first candidate
+
+        Note that the current algorithm will give different results depending
+        on the order of the specs or builders, and also does not consider
+        inheritance hierarchy. Future improvements to this matching algorithm
+        should resolve these discrepancies.
+        """
+        candidates = self._filter_by_name(self._spec_matches, builder)
+        candidates = self._filter_by_type(candidates, builder)
+        if len(candidates) == 0:
+            return None
+        elif len(candidates) == 1:
+            return candidates[0]
+        else:
+            unsatisfied_candidates = self._filter_by_unsatisfied(candidates)
+            if len(unsatisfied_candidates) == 0:
+                return candidates[0]
+            else:
+                return unsatisfied_candidates[0]
+
+    def _filter_by_name(self, candidates, builder):
+        """Returns the candidate specs that either have the same name as the
+        builder or do not specify a name.
+        """
+        def name_is_consistent(spec_matches):
+            spec = spec_matches.spec
+            return spec.name is None or spec.name == builder.name
+
+        return list(filter(name_is_consistent, candidates))
+
+    def _filter_by_type(self, candidates, builder):
+        """Returns the candidate specs which have a data type consistent with
+        the builder's data type.
+        """
+        def compatible_type(spec_matches):
+            spec = spec_matches.spec
+            if isinstance(spec, LinkSpec):
+                validator = self.vmap.get_validator(spec.target_type)
+                spec = validator.spec
+            if spec.data_type is None:
+                return True
+            valid_validators = self.vmap.valid_types(spec.data_type)
+            valid_types = [v.spec.data_type for v in valid_validators]
+            if isinstance(builder, LinkBuilder):
+                dt = builder.builder.attributes.get(spec.type_key())
+            else:
+                dt = builder.attributes.get(spec.type_key())
+            return dt in valid_types
+
+        return list(filter(compatible_type, candidates))
+
+    def _filter_by_unsatisfied(self, candidates):
+        """Returns the candidate specs which are not yet matched against
+        a number of builders which fulfils the quantity for the spec.
+        """
+        def is_unsatisfied(spec_matches):
+            spec = spec_matches.spec
+            n_match = len(spec_matches.builders)
+            if spec.required and n_match == 0:
+                return True
+            if isinstance(spec.quantity, int) and n_match < spec.quantity:
+                return True
+            return False
+
+        return list(filter(is_unsatisfied, candidates))
